//////////////////////////////////////////////////////////////////////////////
//	Copyright (C) 1998 Chris Cant, PHD Computer Consultants Ltd
/////////////////////////////////////////////////////////////////////////////

#include "hiimwdm.h"

/////////////////////////////////////////////////////////////////////////////
// Data global to this module

static ULONG LogLevel;
static PDRIVER_OBJECT pDriverObject;

/////////////////////////////////////////////////////////////////////////////
// Forward declarations of local functions

ULONG
AbcGetStringSize(
	IN PWSTR String
	);

/////////////////////////////////////////////////////////////////////////////
// If the compiler can handle it, make the 
// AbcInitializeEventLog pageable since it can
// only be called from PASSIVE_LEVEL anyway.

#ifdef ALLOC_PRAGMA
#pragma alloc_text( page, AbcInitializeEventLog)
#endif // ALLOC_PRAGMA

/////////////////////////////////////////////////////////////////////////////
//	AbcInitializeEventLog:
//
// Description:
//		This function sets things up so that
//		other parts of the driver can send
//		error messages to the Event Log.
//
// Arguments:
//		Address of Driver object
//
// Return Value:
//		(None)
//--
VOID
AbcInitializeEventLog(
	IN PDRIVER_OBJECT DriverObject
	)
{
	// Remember Driver Object for AbcReportUnexpectedFailure
	pDriverObject = DriverObject;

	LogLevel = DEFAULT_LOG_LEVEL;
	// Log a message saying that logging is enabled.
	AbcReportEvent(
		LOG_LEVEL_DEBUG,
		1,
		ABC_ERRORLOG_INIT,
		(PVOID)DriverObject,
		NULL,					// IRP
		NULL, 0,				// dump data
		NULL, 0);				// strings
}

/////////////////////////////////////////////////////////////////////////////
//	AbcReportEvent:
//
// Description:
//		This function puts a message in
//		the system Event log.
//
// IRQL:
//		< DISPATCH_LEVEL
//
// Arguments:
//		LOG_LEVEL for this message
//		Error message codes
//		Address of Driver or Device object
//		Address of IRP (or NULL)
//		Address of dump data array (or NULL)
//		Count of ULONGs in dump data array (or zero)
//		Address of array of insertion strings
//		Count of insertion strings (or zero)
//		
//
// Return Value:
//		TRUE if the error was logged
//		FALSE if there was a problem

BOOLEAN
AbcReportEvent(
	IN ULONG	MessageLevel,
	IN NTSTATUS	ErrorCode,
	IN ULONG	UniqueErrorValue,
	IN PVOID	IoObject,
	IN PIRP		Irp,
	IN ULONG	DumpData[],
	IN ULONG	DumpDataCount,
	IN PWSTR	Strings[],
	IN ULONG	StringCount
	)
{
	PIO_ERROR_LOG_PACKET Packet;
	PIO_STACK_LOCATION IrpStack;
	PUCHAR pInsertionString;
	UCHAR PacketSize;
	UCHAR StringSize[ ABC_MAX_INSERTION_STRINGS ];
	ULONG i;

	// If we're not logging, or the message
	// is out of range, don't do anything.
/*	if( ( LogLevel == LOG_LEVEL_NONE) ||
		( MessageLevel > LogLevel))
	{
		return TRUE;
	}
*/
	// Start with minimum required packet size
	PacketSize = sizeof(IO_ERROR_LOG_PACKET);

	// Add in any dump data. Remember that the
	// standard error log packet already has one
	// slot in its DumpData array.
	if( DumpDataCount > 0)
		PacketSize += 
				(UCHAR)( sizeof(ULONG) *
							( DumpDataCount - 1 )); 

	// Determine total space needed for any 
	// insertion strings.
	if( StringCount > 0 )
	{
		// Take the lesser of what the caller sent
		// and what this routine can handle
		if( StringCount > ABC_MAX_INSERTION_STRINGS)
			StringCount = ABC_MAX_INSERTION_STRINGS;

		for( i=0; i<StringCount; i++)
		{
			// Keep track of individual string sizes
			StringSize[i] = (UCHAR)AbcGetStringSize( Strings[i]);

			PacketSize += StringSize[i];
		}
	}

	// Try to allocate the packet
	Packet = (PIO_ERROR_LOG_PACKET)IoAllocateErrorLogEntry( IoObject, PacketSize);

	if( Packet==NULL) return FALSE;

	// Fill in standard parts of the packet
	Packet->ErrorCode = ErrorCode;
	Packet->UniqueErrorValue = UniqueErrorValue;

	// If there's an IRP, then the IoObject must
	// be a Device object. In that case, use the
	// IRP and the Device Extension to fill in
	// additional parts of the packet. Otherwise,
	// set these additional fields to zero.
	if( Irp!=NULL)
	{
		IrpStack = IoGetCurrentIrpStackLocation( Irp);
		
		Packet->MajorFunctionCode = IrpStack->MajorFunction;

		Packet->RetryCount = 0;

		Packet->FinalStatus = Irp->IoStatus.Status;

		Packet->SequenceNumber = 0;

		if( IrpStack->MajorFunction == IRP_MJ_DEVICE_CONTROL ||
			IrpStack->MajorFunction == IRP_MJ_INTERNAL_DEVICE_CONTROL)
		{
			Packet->IoControlCode = IrpStack->Parameters.DeviceIoControl.IoControlCode;
		}
		else Packet->IoControlCode = 0;
	}
	else	// No IRP
	{
		Packet->MajorFunctionCode = 0;
		Packet->RetryCount = 0;
		Packet->FinalStatus = 0;
		Packet->SequenceNumber = 0;
		Packet->IoControlCode = 0;
	}

	// Add the dump data
	if( DumpDataCount > 0)
	{
		Packet->DumpDataSize = (USHORT)( sizeof(ULONG) * DumpDataCount);

		for( i=0; i<DumpDataCount; i++)
			Packet->DumpData[i] = DumpData[i];
	}
	else Packet->DumpDataSize = 0;

	// Add the insertion strings
	Packet->NumberOfStrings = (USHORT)StringCount;

	if(StringCount>0)
	{
		// The strings always go just after
		// the DumpData array in the packet
		Packet->StringOffset =
			sizeof(IO_ERROR_LOG_PACKET) +
			( DumpDataCount - 1 ) * sizeof(ULONG);

		pInsertionString = (PUCHAR)Packet + Packet->StringOffset;

		for( i=0; i<StringCount; i++)
		{
			// Add each new string to the end
			// of the existing stuff
			RtlCopyBytes(
				pInsertionString,
				Strings[i],
				StringSize[i]);

			pInsertionString += StringSize[i];
		}
	}

	// Log the message
	IoWriteErrorLogEntry( Packet);

	return TRUE;
}

/////////////////////////////////////////////////////////////////////////////
//	AbcGetStringSize:
//
// Description:
//		This function determines the amount of
//		space needed to hold a NULL-terminated
//		Unicode string.
//
// Arguments:
//		Address of string
//		
//
// Return Value:
//		Size (in bytes) of the string

ULONG
AbcGetStringSize(
	IN PWSTR String
	)
{
	UNICODE_STRING TempString;

	// Use an RTL routine to get the length
	RtlInitUnicodeString( &TempString, String);

	// Size is actually two greater because
	// of the UNICODE_NULL at the end.
	return( TempString.Length + sizeof( WCHAR));
}

/////////////////////////////////////////////////////////////////////////////
